package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.InferenceVariable;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.Substitution;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.WildcardBinding;

public class CaptureBinding extends TypeVariableBinding {
  public TypeBinding lowerBound;
  
  public WildcardBinding wildcard;
  
  public int captureID;
  
  public ReferenceBinding sourceType;
  
  public int start;
  
  public int end;
  
  public ASTNode cud;
  
  TypeBinding pendingSubstitute;
  
  public CaptureBinding(WildcardBinding wildcard, ReferenceBinding sourceType, int start, int end, ASTNode cud, int captureID) {
    super(TypeConstants.WILDCARD_CAPTURE_NAME_PREFIX, wildcard.environment);
    this.wildcard = wildcard;
    this.modifiers = 1073741825;
    this.fPackage = wildcard.fPackage;
    this.sourceType = sourceType;
    this.start = start;
    this.end = end;
    this.captureID = captureID;
    this.tagBits |= 0x2000000000000000L;
    this.cud = cud;
    if (wildcard.hasTypeAnnotations()) {
      org.eclipse.jdt.internal.compiler.lookup.CaptureBinding unannotated = (org.eclipse.jdt.internal.compiler.lookup.CaptureBinding)clone(null);
      unannotated.wildcard = (WildcardBinding)this.wildcard.unannotated();
      this.environment.getUnannotatedType((TypeBinding)unannotated);
      this.id = unannotated.id;
      this.environment.typeSystem.cacheDerivedType((TypeBinding)this, (TypeBinding)unannotated, (TypeBinding)this);
      super.setTypeAnnotations(wildcard.getTypeAnnotations(), wildcard.environment.globalOptions.isAnnotationBasedNullAnalysisEnabled);
      if (wildcard.hasNullTypeAnnotations())
        this.tagBits |= 0x100000L; 
    } else {
      computeId(this.environment);
    } 
  }
  
  protected CaptureBinding(ReferenceBinding sourceType, char[] sourceName, int start, int end, int captureID, LookupEnvironment environment) {
    super(sourceName, null, 0, environment);
    this.modifiers = 1073741825;
    this.sourceType = sourceType;
    this.start = start;
    this.end = end;
    this.captureID = captureID;
  }
  
  public CaptureBinding(org.eclipse.jdt.internal.compiler.lookup.CaptureBinding prototype) {
    super(prototype);
    this.wildcard = prototype.wildcard;
    this.sourceType = prototype.sourceType;
    this.start = prototype.start;
    this.end = prototype.end;
    this.captureID = prototype.captureID;
    this.lowerBound = prototype.lowerBound;
    this.tagBits |= prototype.tagBits & 0x2000000000000000L;
    this.cud = prototype.cud;
  }
  
  public TypeBinding clone(TypeBinding enclosingType) {
    return (TypeBinding)new org.eclipse.jdt.internal.compiler.lookup.CaptureBinding(this);
  }
  
  public char[] computeUniqueKey(boolean isLeaf) {
    StringBuffer buffer = new StringBuffer();
    if (isLeaf) {
      buffer.append(this.sourceType.computeUniqueKey(false));
      buffer.append('&');
    } 
    buffer.append(TypeConstants.WILDCARD_CAPTURE);
    buffer.append(this.wildcard.computeUniqueKey(false));
    buffer.append(this.end);
    buffer.append(';');
    int length = buffer.length();
    char[] uniqueKey = new char[length];
    buffer.getChars(0, length, uniqueKey, 0);
    return uniqueKey;
  }
  
  public String debugName() {
    if (this.wildcard != null) {
      StringBuffer buffer = new StringBuffer(10);
      AnnotationBinding[] annotations = getTypeAnnotations();
      for (int i = 0, length = (annotations == null) ? 0 : annotations.length; i < length; i++) {
        buffer.append(annotations[i]);
        buffer.append(' ');
      } 
      buffer
        .append(TypeConstants.WILDCARD_CAPTURE_NAME_PREFIX)
        .append(this.captureID)
        .append(TypeConstants.WILDCARD_CAPTURE_NAME_SUFFIX)
        .append(this.wildcard.debugName());
      return buffer.toString();
    } 
    return super.debugName();
  }
  
  public char[] genericTypeSignature() {
    if (this.genericTypeSignature == null)
      this.genericTypeSignature = CharOperation.concat(TypeConstants.WILDCARD_CAPTURE, this.wildcard.genericTypeSignature()); 
    return this.genericTypeSignature;
  }
  
  public void initializeBounds(Scope scope, ParameterizedTypeBinding capturedParameterizedType) {
    TypeBinding capturedWildcardBound;
    TypeVariableBinding wildcardVariable = this.wildcard.typeVariable();
    if (wildcardVariable == null) {
      TypeBinding typeBinding2, typeBinding1 = this.wildcard.bound;
      switch (this.wildcard.boundKind) {
        case 1:
          typeBinding2 = typeBinding1.capture(scope, this.start, this.end);
          if (typeBinding1.isInterface()) {
            setSuperClass(scope.getJavaLangObject());
            setSuperInterfaces(new ReferenceBinding[] { (ReferenceBinding)typeBinding2 });
          } else {
            if (typeBinding2.isArrayType() || TypeBinding.equalsEquals(typeBinding2, (TypeBinding)this)) {
              setSuperClass(scope.getJavaLangObject());
            } else {
              setSuperClass((ReferenceBinding)typeBinding2);
            } 
            setSuperInterfaces(Binding.NO_SUPERINTERFACES);
          } 
          setFirstBound(typeBinding2);
          if ((typeBinding2.tagBits & 0x20000000L) == 0L)
            this.tagBits &= 0xFFFFFFFFDFFFFFFFL; 
          break;
        case 0:
          setSuperClass(scope.getJavaLangObject());
          setSuperInterfaces(Binding.NO_SUPERINTERFACES);
          this.tagBits &= 0xFFFFFFFFDFFFFFFFL;
          break;
        case 2:
          setSuperClass(scope.getJavaLangObject());
          setSuperInterfaces(Binding.NO_SUPERINTERFACES);
          this.lowerBound = this.wildcard.bound;
          if ((typeBinding1.tagBits & 0x20000000L) == 0L)
            this.tagBits &= 0xFFFFFFFFDFFFFFFFL; 
          break;
      } 
      return;
    } 
    ReferenceBinding originalVariableSuperclass = wildcardVariable.superclass;
    ReferenceBinding substitutedVariableSuperclass = (ReferenceBinding)Scope.substitute((Substitution)capturedParameterizedType, (TypeBinding)originalVariableSuperclass);
    if (TypeBinding.equalsEquals((TypeBinding)substitutedVariableSuperclass, (TypeBinding)this))
      substitutedVariableSuperclass = originalVariableSuperclass; 
    ReferenceBinding[] originalVariableInterfaces = wildcardVariable.superInterfaces();
    ReferenceBinding[] substitutedVariableInterfaces = Scope.substitute((Substitution)capturedParameterizedType, originalVariableInterfaces);
    if (substitutedVariableInterfaces != originalVariableInterfaces)
      for (int i = 0, length = substitutedVariableInterfaces.length; i < length; i++) {
        if (TypeBinding.equalsEquals((TypeBinding)substitutedVariableInterfaces[i], (TypeBinding)this))
          substitutedVariableInterfaces[i] = originalVariableInterfaces[i]; 
      }  
    TypeBinding originalWildcardBound = this.wildcard.bound;
    switch (this.wildcard.boundKind) {
      case 1:
        capturedWildcardBound = originalWildcardBound.capture(scope, this.start, this.end);
        if (originalWildcardBound.isInterface()) {
          setSuperClass(substitutedVariableSuperclass);
          if (substitutedVariableInterfaces == Binding.NO_SUPERINTERFACES) {
            setSuperInterfaces(new ReferenceBinding[] { (ReferenceBinding)capturedWildcardBound });
          } else {
            int length = substitutedVariableInterfaces.length;
            System.arraycopy(substitutedVariableInterfaces, 0, substitutedVariableInterfaces = new ReferenceBinding[length + 1], 1, length);
            substitutedVariableInterfaces[0] = (ReferenceBinding)capturedWildcardBound;
            setSuperInterfaces(Scope.greaterLowerBound(substitutedVariableInterfaces));
          } 
        } else {
          if (capturedWildcardBound.isArrayType() || TypeBinding.equalsEquals(capturedWildcardBound, (TypeBinding)this)) {
            setSuperClass(substitutedVariableSuperclass);
          } else {
            setSuperClass((ReferenceBinding)capturedWildcardBound);
            if (this.superclass.isSuperclassOf(substitutedVariableSuperclass))
              setSuperClass(substitutedVariableSuperclass); 
          } 
          setSuperInterfaces(substitutedVariableInterfaces);
        } 
        setFirstBound(capturedWildcardBound);
        if ((capturedWildcardBound.tagBits & 0x20000000L) == 0L)
          this.tagBits &= 0xFFFFFFFFDFFFFFFFL; 
        break;
      case 0:
        setSuperClass(substitutedVariableSuperclass);
        setSuperInterfaces(substitutedVariableInterfaces);
        this.tagBits &= 0xFFFFFFFFDFFFFFFFL;
        break;
      case 2:
        setSuperClass(substitutedVariableSuperclass);
        if (TypeBinding.equalsEquals(wildcardVariable.firstBound, (TypeBinding)substitutedVariableSuperclass) || TypeBinding.equalsEquals(originalWildcardBound, (TypeBinding)substitutedVariableSuperclass))
          setFirstBound((TypeBinding)substitutedVariableSuperclass); 
        setSuperInterfaces(substitutedVariableInterfaces);
        this.lowerBound = originalWildcardBound;
        if ((originalWildcardBound.tagBits & 0x20000000L) == 0L)
          this.tagBits &= 0xFFFFFFFFDFFFFFFFL; 
        break;
    } 
    if (scope.environment().usesNullTypeAnnotations())
      evaluateNullAnnotations(scope, null); 
  }
  
  public boolean isCapture() {
    return true;
  }
  
  public boolean isEquivalentTo(TypeBinding otherType) {
    if (equalsEquals((TypeBinding)this, otherType))
      return true; 
    if (otherType == null)
      return false; 
    if (this.firstBound != null && this.firstBound.isArrayType() && 
      this.firstBound.isCompatibleWith(otherType))
      return true; 
    switch (otherType.kind()) {
      case 516:
      case 8196:
        return ((WildcardBinding)otherType).boundCheck((TypeBinding)this);
    } 
    return false;
  }
  
  public boolean isProperType(boolean admitCapture18) {
    if (this.lowerBound != null && !this.lowerBound.isProperType(admitCapture18))
      return false; 
    if (this.wildcard != null && !this.wildcard.isProperType(admitCapture18))
      return false; 
    return super.isProperType(admitCapture18);
  }
  
  public char[] readableName() {
    if (this.wildcard != null) {
      StringBuffer buffer = new StringBuffer(10);
      buffer
        .append(TypeConstants.WILDCARD_CAPTURE_NAME_PREFIX)
        .append(this.captureID)
        .append(TypeConstants.WILDCARD_CAPTURE_NAME_SUFFIX)
        .append(this.wildcard.readableName());
      int length = buffer.length();
      char[] name = new char[length];
      buffer.getChars(0, length, name, 0);
      return name;
    } 
    return super.readableName();
  }
  
  public char[] signableName() {
    if (this.wildcard != null) {
      StringBuffer buffer = new StringBuffer(10);
      buffer
        .append(TypeConstants.WILDCARD_CAPTURE_SIGNABLE_NAME_SUFFIX)
        .append(this.wildcard.readableName());
      int length = buffer.length();
      char[] name = new char[length];
      buffer.getChars(0, length, name, 0);
      return name;
    } 
    return super.readableName();
  }
  
  public char[] shortReadableName() {
    if (this.wildcard != null) {
      StringBuffer buffer = new StringBuffer(10);
      buffer
        .append(TypeConstants.WILDCARD_CAPTURE_NAME_PREFIX)
        .append(this.captureID)
        .append(TypeConstants.WILDCARD_CAPTURE_NAME_SUFFIX)
        .append(this.wildcard.shortReadableName());
      int length = buffer.length();
      char[] name = new char[length];
      buffer.getChars(0, length, name, 0);
      return name;
    } 
    return super.shortReadableName();
  }
  
  public char[] nullAnnotatedReadableName(CompilerOptions options, boolean shortNames) {
    StringBuffer nameBuffer = new StringBuffer(10);
    appendNullAnnotation(nameBuffer, options);
    nameBuffer.append(sourceName());
    if (!this.inRecursiveFunction) {
      this.inRecursiveFunction = true;
      try {
        if (this.wildcard != null) {
          nameBuffer.append("of ");
          nameBuffer.append(this.wildcard.withoutToplevelNullAnnotation().nullAnnotatedReadableName(options, shortNames));
        } else if (this.lowerBound != null) {
          nameBuffer.append(" super ");
          nameBuffer.append(this.lowerBound.nullAnnotatedReadableName(options, shortNames));
        } else if (this.firstBound != null) {
          nameBuffer.append(" extends ");
          nameBuffer.append(this.firstBound.nullAnnotatedReadableName(options, shortNames));
          TypeBinding[] otherUpperBounds = otherUpperBounds();
          if (otherUpperBounds != NO_TYPES)
            nameBuffer.append(" & ..."); 
        } 
      } finally {
        this.inRecursiveFunction = false;
      } 
    } 
    int nameLength = nameBuffer.length();
    char[] readableName = new char[nameLength];
    nameBuffer.getChars(0, nameLength, readableName, 0);
    return readableName;
  }
  
  public TypeBinding withoutToplevelNullAnnotation() {
    if (!hasNullTypeAnnotations())
      return (TypeBinding)this; 
    if (this.wildcard != null && this.wildcard.hasNullTypeAnnotations()) {
      WildcardBinding newWildcard = (WildcardBinding)this.wildcard.withoutToplevelNullAnnotation();
      if (newWildcard != this.wildcard) {
        org.eclipse.jdt.internal.compiler.lookup.CaptureBinding newCapture = (org.eclipse.jdt.internal.compiler.lookup.CaptureBinding)this.environment.getUnannotatedType((TypeBinding)this).clone(null);
        if (newWildcard.hasTypeAnnotations())
          newCapture.tagBits |= 0x200000L; 
        newCapture.wildcard = newWildcard;
        newCapture.superclass = this.superclass;
        newCapture.superInterfaces = this.superInterfaces;
        AnnotationBinding[] newAnnotations = this.environment.filterNullTypeAnnotations(this.typeAnnotations);
        return this.environment.createAnnotatedType((TypeBinding)newCapture, newAnnotations);
      } 
    } 
    return super.withoutToplevelNullAnnotation();
  }
  
  TypeBinding substituteInferenceVariable(InferenceVariable var, TypeBinding substituteType) {
    if (this.pendingSubstitute != null)
      return this.pendingSubstitute; 
    try {
      TypeBinding substitutedWildcard = this.wildcard.substituteInferenceVariable(var, substituteType);
      if (substitutedWildcard != this.wildcard) {
        org.eclipse.jdt.internal.compiler.lookup.CaptureBinding substitute = (org.eclipse.jdt.internal.compiler.lookup.CaptureBinding)clone((TypeBinding)enclosingType());
        substitute.wildcard = (WildcardBinding)substitutedWildcard;
        this.pendingSubstitute = (TypeBinding)substitute;
        if (this.lowerBound != null)
          substitute.lowerBound = this.lowerBound.substituteInferenceVariable(var, substituteType); 
        if (this.firstBound != null)
          substitute.firstBound = this.firstBound.substituteInferenceVariable(var, substituteType); 
        if (this.superclass != null)
          substitute.superclass = (ReferenceBinding)this.superclass.substituteInferenceVariable(var, substituteType); 
        if (this.superInterfaces != null) {
          int length = this.superInterfaces.length;
          substitute.superInterfaces = new ReferenceBinding[length];
          for (int i = 0; i < length; i++)
            substitute.superInterfaces[i] = (ReferenceBinding)this.superInterfaces[i].substituteInferenceVariable(var, substituteType); 
        } 
        return (TypeBinding)substitute;
      } 
      return (TypeBinding)this;
    } finally {
      this.pendingSubstitute = null;
    } 
  }
  
  public void setTypeAnnotations(AnnotationBinding[] annotations, boolean evalNullAnnotations) {
    super.setTypeAnnotations(annotations, evalNullAnnotations);
    if (annotations != Binding.NO_ANNOTATIONS && this.wildcard != null)
      this.wildcard = (WildcardBinding)this.wildcard.environment.createAnnotatedType((TypeBinding)this.wildcard, annotations); 
  }
  
  public TypeBinding uncapture(Scope scope) {
    return (TypeBinding)this.wildcard;
  }
  
  protected TypeBinding[] getDerivedTypesForDeferredInitialization() {
    TypeBinding[] derived = this.environment.typeSystem.getDerivedTypes((TypeBinding)this);
    if (derived.length > 0) {
      int count = 0;
      for (int i = 0; i < derived.length; i++) {
        if (derived[i] != null && (derived[i]).id == this.id)
          derived[count++] = derived[i]; 
      } 
      if (count < derived.length)
        System.arraycopy(derived, 0, derived = new TypeBinding[count], 0, count); 
    } 
    return derived;
  }
  
  public String toString() {
    if (this.wildcard != null) {
      StringBuffer buffer = new StringBuffer(10);
      AnnotationBinding[] annotations = getTypeAnnotations();
      for (int i = 0, length = (annotations == null) ? 0 : annotations.length; i < length; i++) {
        buffer.append(annotations[i]);
        buffer.append(' ');
      } 
      buffer
        .append(TypeConstants.WILDCARD_CAPTURE_NAME_PREFIX)
        .append(this.captureID)
        .append(TypeConstants.WILDCARD_CAPTURE_NAME_SUFFIX)
        .append(this.wildcard);
      return buffer.toString();
    } 
    return super.toString();
  }
}
